"""Main module."""

# Core
import os
import tempfile

# BrainGlobe
from bg_atlasapi import BrainGlobeAtlas
from bg_atlasapi.list_atlases import get_all_atlases_lastversions, get_downloaded_atlases, get_local_atlas_version
from bg_atlasapi.utils import check_internet_connection

# PyImageJ / Scyjava
from scyjava import jimport
import imagej

# JPype
from jpype.types import JString, JArray


def start_imagej(headless: bool = False):
    mode = "headless"
    if not headless:
        mode = "interactive"
    ij = imagej.init(get_java_dependencies(), mode=mode)
    enable_python_hooks(ij)
    add_brainglobe_atlases(ij)

    from scyjava import jimport
    from jpype.types import JString

    # loci.common.DebugTools.enableLogging("OFF");
    DebugTools = jimport('loci.common.DebugTools')
    # DebugTools.enableLogging('OFF')
    DebugTools.enableLogging("INFO");
    # DebugTools.enableLogging("DEBUG");

    import platform
    if platform.system() == 'Windows':
        File = jimport('java.io.File')
        # Now let's set the atlas folder location in a folder with all users access

        AtlasLocationHelper = jimport('ch.epfl.biop.atlas.AtlasLocationHelper')
        directory = os.path.join(os.environ['ProgramData'], 'abba-atlas')

        # create the directory with write access for all users
        try:
            print('Attempt to set ABBA Atlas cache directory to ' + directory)
            os.makedirs(directory, exist_ok=True)
            atlasPath = str(directory)
            AtlasLocationHelper.defaultCacheDir = File(JString(atlasPath))
            print('ABBA Atlas cache directory set to ' + directory)
        except OSError:
            print('ERROR! Could not set ABBA Atlas cache dir')
            # directory already exists ?
            pass
    else:
        print('ERROR! ' + platform.system() + ' OS not supported yet.')
    pass

    if not headless:
        ij.ui().showUI()


def get_java_dependencies():
    """
    Returns the jar files that need to be included into the classpath
    of an imagej object in order to have a functional ABBA app
    these jars should be available in https://maven.scijava.org/
    :return:
    """
    imagej_core_dep = 'net.imagej:imagej:2.9.0'
    imagej_legacy_dep = 'net.imagej:imagej-legacy:0.39.3'
    abba_dep = 'ch.epfl.biop:ImageToAtlasRegister:0.5.3'
    return [imagej_core_dep, imagej_legacy_dep, abba_dep]


def add_brainglobe_atlases(ij):
    # TODO : check connection available or not
    try:
        check_internet_connection()
        available_atlases = get_all_atlases_lastversions()
    except ConnectionError:
        available_atlases_nodict = get_downloaded_atlases()
        available_atlases = dict()
        for atlas in available_atlases_nodict:
            print(atlas)
            available_atlases[atlas] = get_local_atlas_version(atlas)

    AtlasChooserCommand = jimport('ch.epfl.biop.atlas.scijava.AtlasChooserCommand')
    from abba_private.AbbaAtlas import AbbaAtlas  # delayed import because the jvm should be set first
    from jpype import JImplements, JOverride
    Supplier = jimport('java.util.function.Supplier')

    # initialized

    @JImplements(Supplier)
    class AtlasSupplier(object):

        def __init__(self, atlas_name, ij):
            self.atlas_name = atlas_name
            self.ij = ij

        @JOverride
        def get(self):
            bg_atlas = BrainGlobeAtlas(self.atlas_name)
            current_atlas = AbbaAtlas(bg_atlas, self.ij)
            current_atlas.initialize(None, None)
            Abba.opened_atlases[self.atlas_name] = current_atlas
            return current_atlas

    for atlas_name in available_atlases.keys():
        AtlasChooserCommand.registerAtlas(atlas_name, AtlasSupplier(atlas_name, ij))


def enable_python_hooks(ij):
    from jpype import JImplements, JOverride
    PyPostprocessor = jimport('org.scijava.processor.PyPostprocessor')
    Supplier = jimport('java.util.function.Supplier')
    Consumer = jimport('java.util.function.Consumer')
    import logging

    logger = logging.getLogger('ABBAPostprocessor')

    @JImplements(Consumer)
    class IPyWidgetCommandPostProcessor(object):

        @JOverride
        def accept(self, module):
            self.module = module
            outputs = module.getOutputs()
            logger.debug('ABBA post-processing, module ' + str(module))
            print('ABBA post-processing, module ' + str(module))
            for output_key in outputs.keySet():
                if not module.isOutputResolved(output_key):
                    logger.debug('Unresolved output: ' + str(output_key))
                    print('Unresolved output: ' + str(output_key))
                    if str(module.getInfo().getOutput(
                        output_key).getType()) == 'class ch.epfl.biop.atlas.aligner.gui.bdv.BdvMultislicePositionerView':
                        print('MultiSlicePositionerView found')
                        view = module.getOutputs().get(output_key)
                        print(str(view.msp.getAtlas().getName()))
                        if str(view.msp.getAtlas().getName()) == 'Adult Mouse Brain - Allen Brain Atlas V3p1':
                            print('Adult Mouse Brain - Allen Brain Atlas V3p1 detected, installing DeepSlice module')
                            BdvScijavaHelper = jimport('sc.fiji.bdvpg.scijava.BdvScijavaHelper')
                            from jpype import JImplements, JOverride
                            Runnable = jimport('java.lang.Runnable')

                            class DSPlaceHolder(object):
                                def __init__(self, ij):
                                    self.temp_folder = prepare_deepslice_temp_folder(self)
                                    self.ij = ij

                            @JImplements(Runnable)
                            class ExecuteDeepSlicePython(object):

                                def __init__(self, dsp):
                                    self.dsp = dsp

                                @JOverride
                                def run(self):
                                    RegisterSlicesDeepSliceCommand = jimport(
                                        'ch.epfl.biop.atlas.aligner.command.RegisterSlicesDeepSliceCommand')
                                    # Any missing input parameter will lead to a popup window asking the missing argument to the user
                                    self.dsp.ij.command().run(RegisterSlicesDeepSliceCommand, True,
                                                              "image_name_prefix", JString('Section'),
                                                              "mp", view.msp,
                                                              "deepSliceProcessor", self.dsp._run_deep_slice,
                                                              "dataset_folder", JString(self.dsp.temp_folder)
                                                              )

                            BdvScijavaHelper.addActionToBdvHandleMenu(view.getBdvh(),
                                                                      "Align>ABBA - DeepSlice Registration (Python)", 0,
                                                                      ExecuteDeepSlicePython(DSPlaceHolder(ij)))

    @JImplements(Supplier)
    class IPyWidgetCommandPreprocessorSupplier(object):
        @JOverride
        def get(self):
            return IPyWidgetCommandPostProcessor()

    PyPostprocessor.register(IPyWidgetCommandPreprocessorSupplier())
    print('ABBA Python hooks enabled')


def prepare_deepslice_temp_folder(self):
    if not hasattr(self, '_run_deep_slice'):
        from src.abba_python.abba_private.DeepSliceProcessor import DeepSliceProcessor
        self._run_deep_slice = DeepSliceProcessor()

    # TODO : fix potential multiple running instance issues
    temp_folder = tempfile.gettempdir() + '/temp/deepslice/'

    # make sure that the folder exists
    if not os.path.exists(temp_folder):
        os.makedirs(temp_folder)

    # clean folder : remove all previous files
    for filename in os.listdir(temp_folder):
        file_path = os.path.join(temp_folder, filename)
        try:
            if os.path.isfile(file_path) or os.path.islink(file_path):
                os.unlink(file_path)
            # elif os.path.isdir(file_path):
            #    shutil.rmtree(file_path)
        except Exception as e:
            print('Failed to delete %s. Reason: %s' % (file_path, e))
    # print('Temp folder = '+str(temp_folder))
    return temp_folder


class Abba:
    """Abba object which can be used to register sections to a BrainGlobe atlas object
    Parameters
    ----------
    atlas_name :
        Name of the atlas to be used, should be either:
         - available in BrainGlobe
         - or be available in https://github.com/BIOP/ijp-atlas:
            'Adult Mouse Brain - Allen Brain Atlas V3' (default)
            'Rat - Waxholm Sprague Dawley V4'
    ij :
        ImageJ instance, should be reused if you need to open several ABBA instances
        non setting this variable will create an ij instance with minimal dependencies
        If you want to provide your ij instance, make sure that it contains the dependencies
        declared in abba_python.Abba.get_java_dependencies (and all the transitive ones)

        abba_python.ij() returns the current ij instance, which can also be reused in another
        abba_python instance

    slicing_mode :
        should be 'coronal', 'sagittal' or 'horizontal'
        TO IMPROVE : test how well this matches with the BrainGlobe API

    """

    opened_atlases: dict = {}

    def __init__(
        self,
        atlas_name: str = 'Adult Mouse Brain - Allen Brain Atlas V3',
        ij=None,
        slicing_mode: str = 'coronal',  # or sagittal or horizontal
        headless: bool = False,
        enable_jupyter_ui: bool = False
    ):
        if ij is None:
            if headless:
                ij = imagej.init(get_java_dependencies())
                if enable_jupyter_ui:
                    # below : experimental UI, uses ipywidgets instead of ImageJ's swing UI (very partial functionality)
                    from src.abba_python.scijava_python_command import enable_jupyter_ui
                    enable_jupyter_ui()
            else:
                ij = imagej.init(get_java_dependencies(), mode='interactive')
                ij.ui().showUI()
            self.ij = ij
        else:
            print('ij was provided, headless argument ignored')
            self.ij = ij

        # Look in object service to see if the atlas is not already opened by any chance
        # in java TODO

        # or in python
        if atlas_name not in Abba.opened_atlases:
            if atlas_name == 'Adult Mouse Brain - Allen Brain Atlas V3':
                atlas_name = 'Adult Mouse Brain - Allen Brain Atlas V3p1'
            if atlas_name == 'Rat - Waxholm Sprague Dawley V4':
                atlas_name = 'Rat - Waxholm Sprague Dawley V4p1'
            if atlas_name == 'Adult Mouse Brain - Allen Brain Atlas V3p1':
                AllenBrainAdultMouseAtlasCCF2017Command = jimport(
                    'ch.epfl.biop.atlas.mouse.allen.ccfv3p1.command.AllenBrainAdultMouseAtlasCCF2017v3p1Command')
                atlas = ij.command().run(AllenBrainAdultMouseAtlasCCF2017Command, True).get().getOutput("ba")
                Abba.opened_atlases[atlas_name] = atlas
            elif atlas_name == 'Rat - Waxholm Sprague Dawley V4p1':
                WaxholmSpragueDawleyRatV4Command = jimport(
                    'ch.epfl.biop.atlas.rat.waxholm.spraguedawley.v4p1.command.WaxholmSpragueDawleyRatV4p1Command')
                atlas = ij.command().run(WaxholmSpragueDawleyRatV4Command, True).get().getOutput("ba")
                Abba.opened_atlases[atlas_name] = atlas
            else:
                bg_atlas = BrainGlobeAtlas(atlas_name)
                from src.abba_python.abba_private.AbbaAtlas import \
                    AbbaAtlas  # delayed import because the jvm should be correctly
                # initialized
                atlas = AbbaAtlas(bg_atlas, ij)
                atlas.initialize(None, None)
                Abba.opened_atlases[atlas_name] = atlas
                ij.object().addObject(atlas, atlas_name)  # store it in java's object service

        self.atlas = Abba.opened_atlases[atlas_name]
        self.slicing_mode = slicing_mode
        self.atlas_name = atlas_name

        # .. but before : logger, please shut up
        DebugTools = jimport('loci.common.DebugTools')
        DebugTools.enableLogging('OFF')

        # Ok, let's create abba_python's model: mp = multipositioner

        ABBAStartCommand = jimport('ch.epfl.biop.atlas.aligner.command.ABBAStartCommand')  # Command import

        self.mp = ij.command().run(ABBAStartCommand, True,
                                   'slicing_mode', self.slicing_mode,
                                   'ba', self.atlas
                                   ).get().getOutput('mp')

    def get_ij(self):
        """
        Provides the ImageJ instance that can be reused to create another Abba instance
        :return:
            the ImageJ instance used by this Abba instance
        """
        return self.ij

    def show_bdv_ui(self):
        self.ij.ui().showUI()
        """
        Creates and show a BigDataViewer view over this Abba instance
        """
        if not hasattr(self, 'bdv_view'):
            # no bdv view properties : creates a new one
            BdvMultislicePositionerView = jimport('ch.epfl.biop.atlas.aligner.gui.bdv.BdvMultislicePositionerView')
            DefaultBdvSupplier = jimport('sc.fiji.bdvpg.bdv.supplier.DefaultBdvSupplier')
            SerializableBdvOptions = jimport('sc.fiji.bdvpg.bdv.supplier.SerializableBdvOptions')
            bdvh = DefaultBdvSupplier(SerializableBdvOptions()).get()
            self.bdv_view = BdvMultislicePositionerView(self.mp, bdvh)
            if self.atlas_name == 'Adult Mouse Brain - Allen Brain Atlas V3p1':
                if self.slicing_mode == 'coronal':
                    self.install_deepslice_bdv_ui()
                    pass
        else:
            # TODO: make sure it is visible
            pass

    def install_deepslice_bdv_ui(self):

        BdvScijavaHelper = jimport('sc.fiji.bdvpg.scijava.BdvScijavaHelper')

        from jpype import JImplements, JOverride

        Runnable = jimport('java.lang.Runnable')

        @JImplements(Runnable)
        class ExecuteDeepSlicePython(object):

            def __init__(self, abba):
                self.abba = abba

            @JOverride
            def run(self):
                temp_folder = self.abba.prepare_deepslice_temp_folder()
                RegisterSlicesDeepSliceCommand = jimport(
                    'ch.epfl.biop.atlas.aligner.command.RegisterSlicesDeepSliceCommand')
                # Any missing input parameter will lead to a popup window asking the missing argument to the user
                self.abba.ij.command().run(RegisterSlicesDeepSliceCommand, True,
                                           "image_name_prefix", JString('Section'),
                                           "mp", self.abba.mp,
                                           "deepSliceProcessor", self.abba._run_deep_slice,
                                           "dataset_folder", JString(temp_folder)
                                           )

        BdvScijavaHelper.addActionToBdvHandleMenu(self.bdv_view.getBdvh(),
                                                  "Align>ABBA - DeepSlice Registration (Python)", 0,
                                                  ExecuteDeepSlicePython(self))

    def get_bdv_view(self):
        return self.bdv_view

    def prepare_deepslice_temp_folder(self):  # TODO : remove
        if not hasattr(self, '_run_deep_slice'):
            from src.abba_python.abba_private.DeepSliceProcessor import DeepSliceProcessor
            self._run_deep_slice = DeepSliceProcessor()

        # TODO : fix potential multiple running instance issues
        temp_folder = tempfile.gettempdir() + '/temp/deepslice/'

        # make sure that the folder exists
        if not os.path.exists(temp_folder):
            os.makedirs(temp_folder)

        # clean folder : remove all previous files
        for filename in os.listdir(temp_folder):
            file_path = os.path.join(temp_folder, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                # elif os.path.isdir(file_path):
                #    shutil.rmtree(file_path)
            except Exception as e:
                print('Failed to delete %s. Reason: %s' % (file_path, e))
        # print('Temp folder = '+str(temp_folder))
        return temp_folder

    # ------------------------ IMPORT
    def import_from_files(self, filepaths, z_location=0, z_increment=0.02, split_rgb=False):
        """

        :param z_location:
            initial location in mm along the atlas cutting axis
        :param z_increment:
            step in mm between each imported image
        :param split_rgb:
            whether rgb channels should be split in 3 independent RGB channels
            necessary for 16 bits per component RGB images
        :param filepaths:
            file paths of the image to import. Each file should be readable by bio-formats
        :return:
            a Future object: if you call .get(), the request will wait to be finished. If not, the import
            files command is executed asynchronously
        """
        # Let's import the files using Bio-Formats.
        # The list of all commands is accessible here:
        # https://github.com/BIOP/ijp-imagetoatlas/tree/master/src/main/java/ch/epfl/biop/atlas/aligner/command
        ImportSlicesFromFilesCommand = jimport('ch.epfl.biop.atlas.aligner.command.ImportSlicesFromFilesCommand')
        File = jimport('java.io.File')

        # Here we want to import images: check
        # https://github.com/BIOP/ijp-imagetoatlas/blob/master/src/main/java/ch/epfl/biop/atlas/aligner/command/ImportImageCommand.java

        FileArray = JArray(File)
        files = FileArray(len(filepaths))
        i = 0
        for filepath in filepaths:
            file = File(filepath)
            files[i] = file
            i = i + 1

        # Any missing input parameter will lead to a popup window asking the missing argument to the user
        return self.ij.command().run(ImportSlicesFromFilesCommand, True,
                                     "mp", self.mp,
                                     "datasetname", JString('dataset'),
                                     "files", files,
                                     "split_rgb_channels", split_rgb,
                                     "slice_axis_initial_mm", z_location,
                                     "increment_between_slices_mm", z_increment
                                     ).get()

    # ------------------------ SLICE SELECTION
    def select_all_slices(self):
        self.mp.selectSlice(self.mp.getSlices())  # select all

    def deselect_all_slices(self):
        self.mp.deselectSlice(self.mp.getSlices())  # select all

    def get_n_slices(self):
        return self.mp.getSlices().size()

    def select_slices(self, indices):
        for index in indices:
            self.mp.selectSlice(self.mp.getSlices().get(index))  # select the last slice

    # ------------------------ REGISTRATION
    def register_slices_deepslice(self,
                                  channels,
                                  allow_slicing_angle_change=True,
                                  allow_change_slicing_position=True,
                                  maintain_slices_order=True,
                                  affine_transform=True
                                  ):
        # TODO : add option  to rescale brightness/contrast in parameters
        if not self.atlas_name.startswith('Adult Mouse Brain - Allen Brain Atlas V'):
            print('Deep Slice only support the Allen Brain Atlas CCFv3 in coronal slicing mode')
            return

        if not self.slicing_mode == 'coronal':
            print('Deep Slice only support the Allen Brain Atlas CCFv3 in coronal slicing mode')
            return

        temp_folder = self.prepare_deepslice_temp_folder()

        RegisterSlicesDeepSliceCommand = jimport('ch.epfl.biop.atlas.aligner.command.RegisterSlicesDeepSliceCommand')

        # Any missing input parameter will lead to a popup window asking the missing argument to the user
        return self.ij.command().run(RegisterSlicesDeepSliceCommand, True,
                                     "channels", JString(','.join(map(str, channels))),
                                     "image_name_prefix", JString('Section'),
                                     "mp", self.mp,
                                     "allow_slicing_angle_change", allow_slicing_angle_change,
                                     "allow_change_slicing_position", allow_change_slicing_position,
                                     "maintain_slices_order", maintain_slices_order,
                                     "affine_transform", affine_transform,
                                     "deepSliceProcessor", self._run_deep_slice,
                                     "dataset_folder", JString(temp_folder)
                                     ).get()

    def change_display_settings(self, channel_index: int, range_min: float, range_max: float):
        for abba_slice in self.mp.getSlices():
            if abba_slice.isSelected():
                abba_slice.setDisplayRange(channel_index, range_min, range_max)

    def wait_for_end_of_tasks(self):
        self.mp.waitForTasks()

    # ---------------------- AUTOGENERATED CODE
    def close(self):
        ABBACloseCommand = jimport('ch.epfl.biop.atlas.aligner.command.ABBACloseCommand')
        return self.ij.command().run(ABBACloseCommand, True,
                                     'mp', self.mp).get()

    def documentation(self):
        ABBADocumentationCommand = jimport('ch.epfl.biop.atlas.aligner.command.ABBADocumentationCommand')
        return self.ij.command().run(ABBADocumentationCommand, True)

    def forum_help(self):
        ABBAForumHelpCommand = jimport('ch.epfl.biop.atlas.aligner.command.ABBAForumHelpCommand')
        return self.ij.command().run(ABBAForumHelpCommand, True)

    def open_atlas(self,
                   atlastype: str):
        ABBAOpenAtlasCommand = jimport('ch.epfl.biop.atlas.aligner.command.ABBAOpenAtlasCommand')
        return self.ij.command().run(ABBAOpenAtlasCommand, True,
                                     'atlasType', atlastype).get()

    def set_bdv_preferences(self):
        ABBASetBDVPreferencesCommand = jimport('ch.epfl.biop.atlas.aligner.command.ABBASetBDVPreferencesCommand')
        return self.ij.command().run(ABBASetBDVPreferencesCommand, True)

    def state_load(self,
                   state_file):
        ABBAStateLoadCommand = jimport('ch.epfl.biop.atlas.aligner.command.ABBAStateLoadCommand')
        return self.ij.command().run(ABBAStateLoadCommand, True,
                                     'mp', self.mp,
                                     'state_file', state_file).get().getOutput('success')

    def state_save(self,
                   state_file):
        ABBAStateSaveCommand = jimport('ch.epfl.biop.atlas.aligner.command.ABBAStateSaveCommand')
        return self.ij.command().run(ABBAStateSaveCommand, True,
                                     'mp', self.mp,
                                     'state_file', state_file).get().getOutput('success')

    def user_feedback(self):
        ABBAUserFeedbackCommand = jimport('ch.epfl.biop.atlas.aligner.command.ABBAUserFeedbackCommand')
        return self.ij.command().run(ABBAUserFeedbackCommand, True)

    def deep_slice_documentation(self):
        DeepSliceDocumentationCommand = jimport('ch.epfl.biop.atlas.aligner.command.DeepSliceDocumentationCommand')
        return self.ij.command().run(DeepSliceDocumentationCommand, True)

    def export_atlas_to_imagej(self,
                               atlas_channels: str,
                               image_name: str,
                               interpolate: bool,
                               px_size_micron: float):
        ExportAtlasToImageJCommand = jimport('ch.epfl.biop.atlas.aligner.command.ExportAtlasToImageJCommand')
        return self.ij.command().run(ExportAtlasToImageJCommand, True,
                                     'mp', self.mp,
                                     'atlas_channels', atlas_channels,
                                     'image_name', image_name,
                                     'interpolate', interpolate,
                                     'px_size_micron', px_size_micron).get()

    def export_deformation_field_to_imagej(self,
                                           downsampling: int,
                                           max_number_of_iterations: int,
                                           resolution_level: int):
        ExportDeformationFieldToImageJCommand = jimport(
            'ch.epfl.biop.atlas.aligner.command.ExportDeformationFieldToImageJCommand')
        return self.ij.command().run(ExportDeformationFieldToImageJCommand, True,
                                     'mp', self.mp,
                                     'downsampling', downsampling,
                                     'max_number_of_iterations', max_number_of_iterations,
                                     'resolution_level', resolution_level).get()

    def export_registration_to_qupath(self,
                                      erase_previous_file: bool):
        ExportRegistrationToQuPathCommand = jimport(
            'ch.epfl.biop.atlas.aligner.command.ExportRegistrationToQuPathCommand')
        return self.ij.command().run(ExportRegistrationToQuPathCommand, True,
                                     'mp', self.mp,
                                     'erase_previous_file', erase_previous_file).get()

    def export_resampled_slices_to_bdv_source(self,
                                              block_size_x: int,
                                              block_size_y: int,
                                              block_size_z: int,
                                              channels: str,
                                              downsample_x: int,
                                              downsample_y: int,
                                              downsample_z: int,
                                              image_name: str,
                                              interpolate: bool,
                                              margin_z: float,
                                              n_threads: int,
                                              px_size_micron_x: float,
                                              px_size_micron_y: float,
                                              px_size_micron_z: float,
                                              resolution_levels: int):
        ExportResampledSlicesToBDVSourceCommand = jimport(
            'ch.epfl.biop.atlas.aligner.command.ExportResampledSlicesToBDVSourceCommand')
        return self.ij.command().run(ExportResampledSlicesToBDVSourceCommand, True,
                                     'mp', self.mp,
                                     'block_size_x', block_size_x,
                                     'block_size_y', block_size_y,
                                     'block_size_z', block_size_z,
                                     'channels', channels,
                                     'downsample_x', downsample_x,
                                     'downsample_y', downsample_y,
                                     'downsample_z', downsample_z,
                                     'image_name', image_name,
                                     'interpolate', interpolate,
                                     'margin_z', margin_z,
                                     'n_threads', n_threads,
                                     'px_size_micron_x', px_size_micron_x,
                                     'px_size_micron_y', px_size_micron_y,
                                     'px_size_micron_z', px_size_micron_z,
                                     'resolution_levels', resolution_levels).get()

    def export_slices_original_data_to_imagej(self,
                                              channels: str,
                                              resolution_level: int,
                                              verbose: bool):
        ExportSlicesOriginalDataToImageJCommand = jimport(
            'ch.epfl.biop.atlas.aligner.command.ExportSlicesOriginalDataToImageJCommand')
        return self.ij.command().run(ExportSlicesOriginalDataToImageJCommand, True,
                                     'mp', self.mp,
                                     'channels', channels,
                                     'resolution_level', resolution_level,
                                     'verbose', verbose).get()

    def export_slices_to_bdv(self,
                             tag: str):
        ExportSlicesToBDVCommand = jimport('ch.epfl.biop.atlas.aligner.command.ExportSlicesToBDVCommand')
        return self.ij.command().run(ExportSlicesToBDVCommand, True,
                                     'mp', self.mp,
                                     'tag', tag).get()

    def export_slices_to_bdv_json_dataset(self,
                                          file,
                                          tag: str):
        ExportSlicesToBDVJsonDatasetCommand = jimport(
            'ch.epfl.biop.atlas.aligner.command.ExportSlicesToBDVJsonDatasetCommand')
        return self.ij.command().run(ExportSlicesToBDVJsonDatasetCommand, True,
                                     'mp', self.mp,
                                     'file', file,
                                     'tag', tag).get()

    def export_slices_to_imagej(self,
                                channels: str,
                                image_name: str,
                                interpolate: bool,
                                px_size_micron: float):
        ExportSlicesToImageJCommand = jimport('ch.epfl.biop.atlas.aligner.command.ExportSlicesToImageJCommand')
        return self.ij.command().run(ExportSlicesToImageJCommand, True,
                                     'mp', self.mp,
                                     'channels', channels,
                                     'image_name', image_name,
                                     'interpolate', interpolate,
                                     'px_size_micron', px_size_micron).get()

    def export_slices_to_quicknii_dataset(self,
                                          channels: str,
                                          convert_to_8_bits: bool,
                                          convert_to_jpg: bool,
                                          dataset_folder,
                                          image_name: str,
                                          interpolate: bool,
                                          px_size_micron: float):
        ExportSlicesToQuickNIIDatasetCommand = jimport(
            'ch.epfl.biop.atlas.aligner.command.ExportSlicesToQuickNIIDatasetCommand')
        return self.ij.command().run(ExportSlicesToQuickNIIDatasetCommand, True,
                                     'mp', self.mp,
                                     'channels', channels,
                                     'convert_to_8_bits', convert_to_8_bits,
                                     'convert_to_jpg', convert_to_jpg,
                                     'dataset_folder', dataset_folder,
                                     'image_name', image_name,
                                     'interpolate', interpolate,
                                     'px_size_micron', px_size_micron).get()

    def import_slice_from_image_plus(self,
                                     image,
                                     slice_axis_mm: float):
        ImportSliceFromImagePlusCommand = jimport('ch.epfl.biop.atlas.aligner.command.ImportSliceFromImagePlusCommand')
        return self.ij.command().run(ImportSliceFromImagePlusCommand, True,
                                     'mp', self.mp,
                                     'image', image,
                                     'slice_axis_mm', slice_axis_mm).get()

    def import_slices_from_files(self,
                                 datasetname: str,
                                 files,
                                 increment_between_slices_mm: float,
                                 slice_axis_initial_mm: float,
                                 split_rgb_channels: bool):
        ImportSlicesFromFilesCommand = jimport('ch.epfl.biop.atlas.aligner.command.ImportSlicesFromFilesCommand')
        return self.ij.command().run(ImportSlicesFromFilesCommand, True,
                                     'mp', self.mp,
                                     'datasetname', datasetname,
                                     'files', files,
                                     'increment_between_slices_mm', increment_between_slices_mm,
                                     'slice_axis_initial_mm', slice_axis_initial_mm,
                                     'split_rgb_channels', split_rgb_channels).get()

    def import_slices_from_qupath(self,
                                  increment_between_slices_mm: float,
                                  qupath_project,
                                  slice_axis_initial_mm: float):
        ImportSlicesFromQuPathCommand = jimport('ch.epfl.biop.atlas.aligner.command.ImportSlicesFromQuPathCommand')
        return self.ij.command().run(ImportSlicesFromQuPathCommand, True,
                                     'mp', self.mp,
                                     'increment_between_slices_mm', increment_between_slices_mm,
                                     'qupath_project', qupath_project,
                                     'slice_axis_initial_mm', slice_axis_initial_mm).get()

    def register_slices_bigwarp(self,
                                channel_atlas: int,
                                channel_slice: int):
        RegisterSlicesBigWarpCommand = jimport('ch.epfl.biop.atlas.aligner.command.RegisterSlicesBigWarpCommand')
        return self.ij.command().run(RegisterSlicesBigWarpCommand, True,
                                     'mp', self.mp,
                                     'channel_atlas', channel_atlas,
                                     'channel_slice', channel_slice).get()

    def register_slices_edit_last(self,
                                  atlas_channels_csv: str,
                                  reuse_original_channels: bool,
                                  slices_channels_csv: str):
        RegisterSlicesEditLastCommand = jimport('ch.epfl.biop.atlas.aligner.command.RegisterSlicesEditLastCommand')
        return self.ij.command().run(RegisterSlicesEditLastCommand, True,
                                     'mp', self.mp,
                                     'atlas_channels_csv', atlas_channels_csv,
                                     'reuse_original_channels', reuse_original_channels,
                                     'slices_channels_csv', slices_channels_csv).get()

    def register_slices_elastix_affine(self,
                                       channels_atlas_csv: str,
                                       channels_slice_csv: str,
                                       pixel_size_micrometer: float,
                                       background_offset_value_moving: float = 0,
                                       show_imageplus_registration_result: bool = False):
        RegisterSlicesElastixAffineCommand = jimport(
            'ch.epfl.biop.atlas.aligner.command.RegisterSlicesElastixAffineCommand')
        return self.ij.command().run(RegisterSlicesElastixAffineCommand, True,
                                     'mp', self.mp,
                                     'background_offset_value_moving', background_offset_value_moving,
                                     'channels_atlas_csv', channels_atlas_csv,
                                     'channels_slice_csv', channels_slice_csv,
                                     'pixel_size_micrometer', pixel_size_micrometer,
                                     'show_imageplus_registration_result', show_imageplus_registration_result).get()

    def register_slices_elastix_spline(self,
                                       channels_atlas_csv: str,
                                       channels_slice_csv: str,
                                       nb_control_points_x: int,
                                       pixel_size_micrometer: float,
                                       background_offset_value_moving: float = 0,
                                       show_imageplus_registration_result: bool = False):
        RegisterSlicesElastixSplineCommand = jimport(
            'ch.epfl.biop.atlas.aligner.command.RegisterSlicesElastixSplineCommand')
        return self.ij.command().run(RegisterSlicesElastixSplineCommand, True,
                                     'mp', self.mp,
                                     'background_offset_value_moving', background_offset_value_moving,
                                     'channels_atlas_csv', channels_atlas_csv,
                                     'channels_slice_csv', channels_slice_csv,
                                     'nb_control_points_x', nb_control_points_x,
                                     'pixel_size_micrometer', pixel_size_micrometer,
                                     'show_imageplus_registration_result', show_imageplus_registration_result).get()

    def register_slices_remove_last(self):
        RegisterSlicesRemoveLastCommand = jimport('ch.epfl.biop.atlas.aligner.command.RegisterSlicesRemoveLastCommand')
        return self.ij.command().run(RegisterSlicesRemoveLastCommand, True,
                                     'mp', self.mp).get()

    def rotate_slices(self,
                      angle_degrees: float,
                      axis_string: str):
        RotateSlicesCommand = jimport('ch.epfl.biop.atlas.aligner.command.RotateSlicesCommand')
        return self.ij.command().run(RotateSlicesCommand, True,
                                     'mp', self.mp,
                                     'angle_degrees', angle_degrees,
                                     'axis_string', axis_string).get()

    def set_slices_display_range(self,
                                 channels_csv: str,
                                 display_max: float,
                                 display_min: float):
        """

        Parameters
        ----------
        channels_csv
            ch
        display_max
        display_min

        Returns
        -------

        """
        SetSlicesDisplayRangeCommand = jimport('ch.epfl.biop.atlas.aligner.command.SetSlicesDisplayRangeCommand')
        return self.ij.command().run(SetSlicesDisplayRangeCommand, True,
                                     'mp', self.mp,
                                     'channels_csv', channels_csv,
                                     'display_max', display_max,
                                     'display_min', display_min).get()

    def set_slices_thickness(self,
                             thickness_in_micrometer: float):
        SetSlicesThicknessCommand = jimport('ch.epfl.biop.atlas.aligner.command.SetSlicesThicknessCommand')
        return self.ij.command().run(SetSlicesThicknessCommand, True,
                                     'mp', self.mp,
                                     'thickness_in_micrometer', thickness_in_micrometer).get()

    def set_slices_thickness_match_neighbors(self):
        SetSlicesThicknessMatchNeighborsCommand = jimport(
            'ch.epfl.biop.atlas.aligner.command.SetSlicesThicknessMatchNeighborsCommand')
        return self.ij.command().run(SetSlicesThicknessMatchNeighborsCommand, True,
                                     'mp', self.mp).get()



import numpy as np
from src.abba_python.deepslice.utilities import QuickNII_functions
import glob
import pandas as pd
def points_to_3d(points, plane):
    resolution = points[['width', 'height']].values.astype(np.float64)[0]
    points = points[['X', 'Y']].values.astype(np.float64)
    ## Scale points by the resolution (numbers will now be between 0 and 1
    points/=resolution
    O_plane = plane[['ox','oy','oz']].values.astype(np.float64)
    X_plane = plane[['ux','uy','uz']].values.astype(np.float64)
    Y_plane = plane[['vx','vy','vz']].values.astype(np.float64)
    ## Since we have the vectors U&V and the fractional cell coordinates
    ## We can convert these onto the points on the plane
    ## Should probably be vectorised
    X_dim   = [X_plane * points[i, 0] for i in range(len(points))]
    Y_dim   = [Y_plane * points[i, 1] for i in range(len(points))]
    points  = np.sum((X_dim,Y_dim), axis=0)
    ##Now we add the origin to these coordinates to transform them into our alignment space
    points += O_plane
    points  = points.reshape((len(points), 3))
    return points

def analyse_slice(point_name, plane):
    image_name = point_name.split('/')[-1][:-4]
    raw_names  = [name[0].split('/')[-1] for name in plane["Filenames"].str.split('.')]
    index      = [i==image_name for i in raw_names]
    plane      = plane[index]
    points     = pd.read_csv(point_name)
    points     = points_to_3d(points, plane)
    return points

def analyse_brain(plane_file, points_folder = 'points/', save_csv=False):
    plane =  QuickNII_functions.XML_to_csv(plane_file)
    cells = np.empty((0,3))
    for points in glob.glob(points_folder+'/*.csv'):
        cells = np.vstack((cells, analyse_slice(points, plane)))
    cells = pd.DataFrame(cells, columns = ['X', 'Y', 'Z'])
    cells.to_hdf(plane_file[:-4]+'Whole_Brain_Cell_Count.h5')
    if save_csv:
        cells.to_csv(plane_file[:-4]+'Whole_Brain_Cell_Count.csv')





__author__ = """Nicolas Chiaruttini"""
__version__ = "0.1.0-SNAPSHOT"

# Inner classes, do not access and use them directly
# Used internally by the Abba package

# from abba_python.abba_private.AbbaAtlasNode import AbbaAtlasNode



from IPython.core.magic import register_cell_magic
from jupyter_ui_poll import ui_events
from threading import Thread
from IPython.core.interactiveshell import InteractiveShell
import time


@register_cell_magic('cell_with_modal_ui')
def cell_with_modal_ui(line, cell):
    shell = InteractiveShell.instance()

    def f():
        shell.run_cell(cell)

    cell_runner = Thread(target=f)
    cell_runner.start()
    with ui_events() as poll:
        while cell_runner.is_alive():
            poll(10)  # React to UI events (upto 10 at a time)
            time.sleep(0.1)



from scyjava import jimport
from jpype.types import JObject, JClass

import logging

PyCommandBuilder = jimport('org.scijava.command.PyCommandBuilder')
PyParameterBuilder = jimport('org.scijava.command.PyParameterBuilder')

# Decorator that registers a python CLASS containing a method named "run" as a Scijava Command
#
# This uses PyCommandBuilder which is in the java repo ch.epfl.biop:pyimagej-scijava-command
# PyCommandBuilder allows to build a Command fully programmatically without using any
# java annotation as java annotations are needed for 'easy' Scijava Commands definition
# but these are not completely supported in JPype:
# cf https://github.com/jpype-project/jpype/issues/940
#
# Example of registering a Scijava Command via the @ScijavaCommand decorator:
# ------------------------------------------
# @ScijavaCommand(context=ij.context(),  # ij context needed
#                 name='pyCommand.HelloCommand')
# @ScijavaInput('name', JString,
#               label='Name :', description='Please enter your name')
# @ScijavaInput('familiar', JBoolean,
#               label='Familiar :', description='Hi or Hello ?')
# @ScijavaOutput('greetings', JString)
# class HelloCommand:
#
#     def run(self):
#         if (self.familiar):
#             self.greetings = 'Hi ' + str(self.name) + '!'
#         else:
#             self.greetings = 'Hello my dear ' + str(self.name) + '.'
#         print(self.greetings)
# ------------------------------------------
#
# Note: this way of defining a command is probably not ideal if this has to be used from the python side also
#
# Because it's a preliminary work, this decorator prints a lot of stuff in the process
#
# TODO: functools ??

logger = logging.getLogger('ScijavaCommand')

def ScijavaCommand(name, context):
    logger.info("- Registering scijava command " + name)

    def registerCommand(func):

        # This class will be registered as a SciJava Command
        builder = PyCommandBuilder()  # Java PyCommandBuilder

        # The name of the command - to avoid name conflicts, consider a 'virtual' class name with its package
        builder = builder.name(name).label(name)

        # Register all inputs
        if hasattr(func, 'scijava_inputs'):
            logger.debug('- Inputs')
            for scijava_input_name, scijava_input_properties in func.scijava_inputs.items():
                logger.debug('\t'+ scijava_input_name+ ' : '+ str(scijava_input_properties['scijava_class']))
                builder = builder.input(scijava_input_name,
                                        scijava_input_properties['scijava_class'],
                                        scijava_input_properties['scijava_parameter'])
                setattr(func, scijava_input_name, None)  # declares empty input field
            logger.debug('Inputs registered')
        else:
            logger.debug('- No input')

        # Register all outputs
        if hasattr(func, 'scijava_outputs'):
            logger.debug('- Outputs')
            for scijava_output_name, scijava_output_properties in func.scijava_outputs.items():
                logger.debug('\t'+ scijava_output_name+ ' : '+ str(scijava_output_properties['scijava_class']))
                builder = builder.output(scijava_output_name,
                                         scijava_output_properties['scijava_class'],
                                         scijava_output_properties['scijava_parameter'])
                setattr(func, scijava_output_name, None)  # declares empty input field
            logger.debug('Outputs registered')
        else:
            logger.debug('- No Output')

        # Wraps the run function - takes kwargs as input, returns outputs
        def wrapped_run(inner_kwargs):
            inner_object = func()
            logger.debug('Settings inputs...')
            if hasattr(func, 'scijava_inputs'):
                for input_name in func.scijava_inputs.keys():
                    setattr(inner_object, input_name, inner_kwargs[input_name])
            logger.debug('Inputs set.')
            logger.debug('Running scijava command: ' + name)
            inner_object.run()
            logger.debug(name + ' command execution done.')
            logger.debug('Fetching outputs...')
            outputs = {}
            if hasattr(func, 'scijava_outputs'):
                for output_name in func.scijava_outputs.keys():
                    outputs[output_name] = getattr(inner_object, output_name)  # gets outputs
            logger.debug('Outputs set.')
            return JObject(outputs, JClass('java.util.Map'))  # Returns output as a java HashMap

        # Sets the function in PyCommandBuilder:
        # Function<Map<String, Object>, Map<String, Object>> command
        builder = builder.function(wrapped_run)

        # Effectively registers this command to the ij context
        builder.create(context)
        return func

    return registerCommand


def ScijavaInput(scijava_input_name, scijava_class, **kwargs):
    logger.debug("- Registering scijava input "+scijava_input_name)
    def registerInput(func):

        if not hasattr(func, 'scijava_inputs'):
            setattr(func, 'scijava_inputs', dict())

        if scijava_input_name in func.scijava_inputs:
            raise Exception("Error, two inputs have the same name " + str(scijava_input_name))

        scijava_parameter = PyParameterBuilder.input()

        for key, value in kwargs.items():
            getattr(scijava_parameter, key)(value) # builder pattern powa!

        func.scijava_inputs[scijava_input_name] = \
            {'scijava_class': scijava_class,
             'scijava_parameter': scijava_parameter.get()}

        return func

    return registerInput


def ScijavaOutput(scijava_output_name, scijava_class, **kwargs):
    logger.debug("- Registering scijava output "+scijava_output_name)

    def registerOutput(func):

        if not hasattr(func, 'scijava_outputs'):
            setattr(func, 'scijava_outputs', dict())

        if scijava_output_name in func.scijava_outputs:
            raise Exception("Error, two outputs have the same name " + str(scijava_output_name))

        scijava_parameter = PyParameterBuilder.output()

        for key, value in kwargs.items():
            getattr(scijava_parameter, key)(value) # builder pattern powa!

        func.scijava_outputs[scijava_output_name] = \
            {'scijava_class': scijava_class,
             'scijava_parameter': scijava_parameter.get()}

        return func

    return registerOutput



# core dependencies
import time

from abba_python.abba_python import enable_python_hooks, add_brainglobe_atlases
# in order to wait for a jvm shutdown
import jpype
import imagej

import os

# THIS FILE SETS MANY PATHS EXPLICITLY WHEN ABBA IS INSTALLED FROM THE INSTALLER!
# IF YOU WANT TO RUN ABBA FROM PYTHON, TRY run-abba.py first!

if __name__ == '__main__':
    os.path.dirname(os.getcwd())
    # In ABBA PYthon, Fiji.app is in the parent directory of this script
    fiji_app_path = str(os.path.join(os.path.dirname(os.getcwd()), 'Fiji.app'))
    ij = imagej.init(fiji_app_path, mode="interactive")

    ij.ui().showUI()
    enable_python_hooks(ij)
    add_brainglobe_atlases(ij)

    # Set Elastix Path:
    # File ch.epfl.biop.wrappers.elastix.Elastix exePath
    # File ch.epfl.biop.wrappers.transformix.Transformix exePath
    from scyjava import jimport
    from jpype.types import JString

    # Java class imports

    DebugTools = jimport('loci.common.DebugTools')
    File = jimport('java.io.File')
    # DebugTools.enableLogging('OFF')
    DebugTools.enableLogging("INFO");
    # DebugTools.enableLogging("DEBUG");

    import platform
    if platform.system() == 'Windows':
        elastixPath = str(os.path.join(os.path.dirname(os.getcwd()), 'win', 'elastix-5.0.1-win64', 'elastix.exe'))
        transformixPath = str(os.path.join(os.path.dirname(os.getcwd()), 'win', 'elastix-5.0.1-win64', 'transformix.exe'))

        Elastix = jimport('ch.epfl.biop.wrappers.elastix.Elastix')
        Elastix.exePath = JString(str(elastixPath))
        Elastix.setExePath(File(JString(str(elastixPath))))
        Transformix = jimport('ch.epfl.biop.wrappers.transformix.Transformix')
        Transformix.exePath = JString(str(transformixPath))
        Transformix.setExePath(File(JString(str(transformixPath))))

        # Now let's set the atlas folder location in a folder that all users can access

        AtlasLocationHelper = jimport('ch.epfl.biop.atlas.AtlasLocationHelper')
        directory = os.path.join(os.environ['ProgramData'], 'abba-atlas')

        # create the directory with write access for all users
        try:
            print('Attempt to set ABBA Atlas cache directory to ' + directory)
            # //os.mkdir(directory)
            os.makedirs(directory, exist_ok=True)
            atlasPath = str(directory)
            AtlasLocationHelper.defaultCacheDir = File(JString(atlasPath))
            print('ABBA Atlas cache directory set to ' + directory)
        except OSError:
            print('ERROR! Could not set ABBA Atlas cache dir')
            # directory already exists ?
            pass
    else:
        print('ERROR! '+platform.system()+' OS not supported yet.')

    # --

    # Wait for the JVM to shut down
    while jpype.isJVMStarted():
        time.sleep(1)

    print("JVM has shut down")



# core dependencies
import time
from src.abba_python.abba_python import start_imagej

# in order to wait for a jvm shutdown
import jpype


if __name__ == '__main__':

    # MAC ISSUE
    # https: // github.com / imagej / pyimagej / issues / 23
    # -- FOR DEBUGGING
    # import imagej.doctor
    # imagej.doctor.checkup()
    # imagej.doctor.debug_to_stderr()
    # -- Atlas
    # Any brainglobe atlas can be used
    # show_atlases()
    # abba_python = Abba("azba_zfish_4um", slicing_mode='sagittal', headless=True)  # or any other brainglobe atlas

    # -- HEADLESS
    # abba_python = Abba('Adult Mouse Brain - Allen Brain Atlas V3', headless=True)  # or any other brainglobe atlas
    # --

    # -- NOT HEADLESS
    # abba = Abba('Adult Mouse Brain - Allen Brain Atlas V3')
    # abba.show_bdv_ui()  # creates and show a bdv view

    start_imagej(headless=False)
    # Wait for the JVM to shut down
    while jpype.isJVMStarted():
        time.sleep(1)

    print("JVM has shut down")



import threading
import ipywidgets as widgets
from IPython.display import display
from scyjava import jimport
from jpype import JImplements, JOverride
from jpype.types import JString, JBoolean, JDouble, JInt, JFloat

from ipyfilechooser import FileChooser
import logging

logger = logging.getLogger('ScijavaJupyterUI')

Consumer = jimport('java.util.function.Consumer')
Supplier = jimport('java.util.function.Supplier')
JFile = jimport('java.io.File')
PyPreprocessor = jimport('org.scijava.processor.PyPreprocessor')


def enable_jupyter_ui():
    PyPreprocessor = jimport('org.scijava.processor.PyPreprocessor')
    PyPreprocessor.register(IPyWidgetCommandPreprocessorSupplier())
    print('Scijava jupyter ui enabled')


@JImplements(Consumer)
class IPyWidgetCommandPreprocessor(object):

    @JOverride
    def accept(self, module):
        self.module = module
        inputs = module.getInputs()
        self.all_widgets = dict()
        logger.debug('Jupyter pre-processing, module ' + str(module))
        for input_key in inputs.keySet():
            if not module.isInputResolved(input_key):
                logger.debug('Unresolved input: ' + str(input_key))
                current_widget = get_jupyter_widget(module, input_key)
                if current_widget is not None:
                    logger.debug('Widget acquired for input ' + str(input_key))
                    self.all_widgets[input_key] = current_widget
                else:
                    # maybe log something or put a warning
                    logger.debug('Widget not found for input ' + str(input_key))
                    pass

        if len(self.all_widgets) != 0:
            list_of_widgets = [x.get_widget() for key, x in self.all_widgets.items()]

            # Creates an OK button, and its associated event
            user_has_clicked = threading.Event()
            ok_button = widgets.Button(description='OK')
            ok_button.on_click(lambda _: user_has_clicked.set())
            list_of_widgets.append(ok_button)
            logger.debug('Display widgets module ' + str(module))
            display(widgets.VBox(list_of_widgets))
            logger.debug('Waiting for click')

            # Wait for button click
            user_has_clicked.wait()
            logger.debug('button clicked')

            # the user has clicked
            # retrieve contents of all widgets:
            for key, w in self.all_widgets.items():
                logger.debug('Resolving input ' + str(key))
                module.setInput(key, w.get_value())
                module.resolveInput(key)
                logger.debug('Input ' + str(key) + " resolved")


@JImplements(Supplier)
class IPyWidgetCommandPreprocessorSupplier(object):
    @JOverride
    def get(self):
        return IPyWidgetCommandPreprocessor()


booleanClasses = ['class java.lang.Boolean', 'boolean']
intClasses = ['class java.lang.Integer', 'int']
floatClasses = ['class java.lang.Float', 'float']
doubleClasses = ['class java.lang.Double', 'double']


def getLabel(module, input_key):
    label = module.getInfo().getInput(input_key).getLabel()
    if (label is None) or (label == ''):
        label = str(input_key)
    else:
        label = str(label)
    return label


def get_jupyter_widget(module, input_key):
    if str(module.getInfo().getInput(input_key).getType()) == 'class java.lang.String':
        return JupyterTextWidget(module, input_key)

    if str(module.getInfo().getInput(input_key).getType()) in booleanClasses:
        return JupyterToggleWidget(module, input_key)

    if str(module.getInfo().getInput(input_key).getType()) in intClasses:
        return JupyterIntWidget(module, input_key)

    if str(module.getInfo().getInput(input_key).getType()) in floatClasses:
        return JupyterFloatWidget(module, input_key)

    if str(module.getInfo().getInput(input_key).getType()) in doubleClasses:
        return JupyterDoubleWidget(module, input_key)

    if str(module.getInfo().getInput(input_key).getType()) == 'class java.io.File':
        return JupyterFileWidget(module, input_key)

    print("Unsupported widget for type " + str(module.getInfo().getInput(input_key).getType()))

    return None


class JupyterInputWidget:
    def __init__(self, module, input_key):
        pass

    def get_widget(self):
        return self.widget

    def get_value(self):
        return None


class JupyterTextWidget(JupyterInputWidget):
    def __init__(self, module, input_key):
        self.widget = widgets.Text(
            value=str(module.getInput(input_key)),
            placeholder=str(module.getInfo().getInput(input_key).getDescription()),
            description=getLabel(module, input_key),
            disabled=False
        )

    def get_value(self):
        return JString(self.widget.value)


class JupyterToggleWidget(JupyterInputWidget):
    def __init__(self, module, input_key):
        self.widget = widgets.Checkbox(
            value=bool(module.getInput(input_key)),
            description=getLabel(module, input_key),
            disabled=False,
            indent=True
        )

    def get_value(self):
        return JBoolean(self.widget.value)


class JupyterIntWidget(JupyterInputWidget):
    def __init__(self, module, input_key):
        min_value = module.getInfo().getInput(input_key).getMinimumValue()
        max_value = module.getInfo().getInput(input_key).getMaximumValue()
        step_size = module.getInfo().getInput(input_key).getStepSize()

        if (min_value is not None) and (max_value is not None):

            if step_size is None:
                step_size = 1
            else:
                step_size = step_size.intValue()

            self.widget = widgets.BoundedIntText(
                value=int(module.getInput(input_key)),
                min=int(min_value),
                max=int(max_value),
                step=int(step_size),
                description=getLabel(module, input_key),
                disabled=False
            )
        else:
            self.widget = widgets.IntText(
                value=int(module.getInput(input_key)),
                description=getLabel(module, input_key),
                disabled=False
            )

    def get_value(self):
        return JInt(self.widget.value)


class JupyterFloatWidget(JupyterInputWidget):
    def __init__(self, module, input_key):

        min_value = module.getInfo().getInput(input_key).getMinimumValue()
        max_value = module.getInfo().getInput(input_key).getMaximumValue()
        step_size = module.getInfo().getInput(input_key).getStepSize()

        if (min_value is not None) and (max_value is not None):
            if step_size is None:
                step_size = 1
            else:
                step_size = step_size.intValue()

            self.widget = widgets.FloatText(
                value=float(module.getInput(input_key)),
                min=float(min_value),
                max=float(max_value),
                step=float(step_size),
                description=getLabel(module, input_key),
                disabled=False
            )
        else:
            self.widget = widgets.BoundedFloatText(
                value=float(module.getInput(input_key)),
                description=getLabel(module, input_key),
                disabled=False
            )

    def get_value(self):
        return JFloat(self.widget.value)


class JupyterDoubleWidget(JupyterFloatWidget):
    def __init__(self, module, input_key):
        JupyterFloatWidget.__init__(self, module, input_key)

    def get_value(self):
        return JDouble(self.widget.value)


class JupyterFileWidget(JupyterInputWidget):
    def __init__(self, module, input_key):

        styles = module.getInfo().getInput(input_key).getWidgetStyle().replace(" ", "").split(',')

        if 'save' in styles:
            self.save = True
            self.widget = widgets.Text(
                value=str(module.getInput(input_key)),
                placeholder=str(module.getInfo().getInput(input_key).getDescription()),
                description=getLabel(module, input_key),
                disabled=False
            )

        else:
            self.save = False

            # Create and display a FileChooser widget
            fc = FileChooser(select_default=True)
            # Change the title
            fc.title = getLabel(module, input_key)

            if 'directory' in styles:
                # Switch to folder-only mode
                fc.show_only_dirs = True
                if module.getInput(input_key) is not None:
                    fc.default_path = str(module.getInput(input_key))
                    fc.reset()
            else:
                if module.getInput(input_key) is not None:
                    fc.default_path = str(module.getInput(input_key).getParent())
                    fc.default_filename = str(module.getInput(input_key).getName())
                    fc.reset()

            matching_extensions = [s for s in styles if 'extensions:' in s]
            if len(matching_extensions) == 1:
                all_extensions = matching_extensions[0].split(":")[1].split("/")
                # Set multiple file filter patterns (uses https://docs.python.org/3/library/fnmatch.html)
                fc.filter_pattern = ['*.' + str(ext) for ext in all_extensions]

            self.widget = fc

    def get_value(self):
        if self.save:
            return JFile(JString(self.widget.value))
        else:
            return JFile(JString(self.widget.selected))  # TODO : fix casting



__author__ = """nicolas_chiaruttini"""
__version__ = "0.1.0-SNAPSHOT"

#from scijava_python_command.MagicJupyter import run_in_another_thread
#from scijava_python_command.ScijavaJupyterUI import enable_jupyter_ui
#from scijava_python_command.ScijavaCommand import ScijavaCommand, ScijavaInput, ScijavaOutput



from scyjava import jimport
from ..deepslice import DeepSlice
from jpype import JImplements, JOverride
from jpype.types import JString
import os

import urllib.request
from tqdm import tqdm
from ..deepslice.DeepSlice import get_deepslice_path

# Import deepslice and make the function for the ABBA command
Function = jimport('java.util.function.Function')
File = jimport('java.io.File')


# ----- Download with progress bar, cf https://stackoverflow.com/questions/15644964/python-progress-bar-and-downloads
class DownloadProgressBar(tqdm):
    def update_to(self, b=1, bsize=1, tsize=None):
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)


def download_url(url, output_path):
    with DownloadProgressBar(unit='B', unit_scale=True,
                             miniters=1, desc=url.split('/')[-1]) as t:
        urllib.request.urlretrieve(url, filename=output_path, reporthook=t.update_to)


# ------------

def check_model_is_present():
    path_to_model = get_deepslice_path() + "/NN_weights"
    if not os.path.isdir(path_to_model):
        print('DeepSlice model not present, 240 Mb will be downloaded ')
        os.makedirs(path_to_model)
    model_files = ['/Allen_Mixed_Best.h5',
                   '/Synthetic_data_final.hdf5',
                   '/xception_weights_tf_dim_ordering_tf_kernels.h5']

    deepslice_model_url = 'https://github.com/PolarBean/DeepSlice/raw/master/NN_weights'

    for file in model_files:
        if not os.path.exists(path_to_model + file):
            target_url = deepslice_model_url + file
            print('Missing DeepSlice model file ' + file + '. Downloading (80Mb)...')
            download_url(target_url, path_to_model + file)


@JImplements(Function)
class DeepSliceProcessor:

    @JOverride
    def apply(self, folder):
        check_model_is_present()
        model = DeepSlice()
        model.Build()
        model.predict(image_dir=str(folder.getParent()))
        out = File(folder, JString('results'))
        model.Save_Results(out.getAbsolutePath())
        return File(folder, JString('results.xml'))



import random
from statistics import mean
import numpy as np
import math
from PIL import Image
import glob
image_list = []
# a 2-dimensional plane can be fully describable if you know any three points that lie on its surface


def find_plane_equation(plane):
    '''
    Finds the plane equation of a plane
    :param plane: the plane to find the equation of
    :type plane: :any:`numpy.ndarray`
    :returns: the normal vector of the plane and the constant k
    :rtype: :any:`numpy.ndarray`, float
    '''
    a, b, c = np.array(plane[0:3], dtype=np.float64), np.array(plane[3:6], dtype=np.float64), np.array(plane[6:9],
                                                                                                       dtype=np.float64)
    cross = np.cross(b, c)
    cross /= 9
    k = -((a[0] * cross[0]) + (a[1] * cross[1]) + (a[2] * cross[2]))
    return (cross, k)

def get_angle(inp, cross, k, direction):
    # inp is the input plane, represented by 3 xzy sets
    # cross and k is the normal vector of the plane
    # Direction defines whether we want the mediolateral or dorsaventral angle
    section = inp.copy()
    # transform vector into absolute coordinates
    for i in range(3):
        section[i + 3] += section[i]
        section[i + 6] += section[i]
    if direction == 'ML':
        # original xzy point
        a = section[0:2]
        # calculate a point which differs from this point only in the x dimension
        # to do this we use the plane equation which tells us the position of every point on the plane
        linear_point = (((section[0] - 100) * cross[0]
                         ) + ((section[2]) * cross[2])) + k
        # this tells us the depth of that point which differs in x dimension but lies on the same plane
        depth = -(linear_point / cross[1])
        b = np.array((section[0] - 100, depth))
        c = b + [100, 0]

        
    if direction == 'DV':
        a = section[1:3]
        linear_point = (((section[0]) * cross[0]) +
                        ((section[2] - 100) * cross[2])) + k
        depth = -(linear_point / cross[1])
        b = np.array((depth, section[2] - 100))
        c = b + [0, 100]
    ba = a - b
    bc = c - b
    cosine_angle = np.dot(ba, bc) / (np.linalg.norm(ba) * np.linalg.norm(bc))
    # This looks redundant, needs to be tested
    angle = np.arccos(cosine_angle)
    angle = np.degrees(angle)
    if direction == 'ML':
        if b[1] > a[1]:
            angle *= -1
    if direction == 'DV':
        if b[0] < a[0]:
            angle *= -1
    return (angle)


def rotation_around_axis(axis, angle):
    '''
    Generates a 3x3 rotation matrix using the Euler-Rodrigues formula
    following the definition here:
    https://en.wikipedia.org/wiki/Euler%E2%80%93Rodrigues_formula.
    :param axis: the axis around which to rotate as a vector of length 3
                 (no normalisation required)
    :type axis: array like
    :param angle: the angle in radians to rotate
    :type angle: float
    :returns: the rotation matrix
    :rtype: a 3x3 :any:`numpy.ndarray`
    '''
    angle = np.radians(angle)
    axis = axis / np.linalg.norm(axis)

    a = math.cos(angle / 2.0)
    b, c, d = axis * math.sin(angle / 2.0)
    aa, bb, cc, dd = a * a, b * b, c * c, d * d
    bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d

    return np.array([[aa + bb - cc - dd, 2 * (bc - ad), 2 * (bd + ac)],
                     [2 * (bc + ad), aa + cc - bb - dd, 2 * (cd - ab)],
                     [2 * (bd - ac), 2 * (cd + ab), aa + dd - bb - cc]])


def make_gaussian_weights(mini, maxi):
    '''
    Generates a list of weights for a gaussian distribution
    :param mini: the minimum value of the distribution
    :type mini: float
    :param maxi: the maximum value of the distribution
    :type maxi: float
    :returns: a list of weights
    :rtype: list
    '''
    weights = []
    center = mean((mini, maxi))
    quartile = (maxi-mini)/4
    while len(weights) < ((maxi-mini)/2):
        weights.append(random.gauss(center, quartile))
    weights = [max(x, mini) for x in weights]
    weights = [min(x, maxi) for x in weights]
    weights = [x-center if x > center else x for x in weights]
    weights = sorted(weights)
    weights_rev = sorted(weights, reverse=True)
    weights.extend(weights_rev)
    return [x/max(weights) for x in weights]


def get_axis(m, translation_vector, direction, plane_of_section=None, atlas='AMBA'):
    '''
    :param m: the matrix to rotate
    :type m: 3x3 :any:`numpy.ndarray`
    :param translation_vector: the translation vector to apply
    :type translation_vector: 3x1 :any:`numpy.ndarray`
    :param direction: the direction of the rotation
    :type direction: string
    :param plane: the plane to rotate around
    :type plane: string
    :returns: the axis of rotation
    :rtype: 3x1 :any:`numpy.ndarray`
    '''
    # find the plane equation for a set of QuickNII coordinates
    cross, k = find_plane_equation(m)
    if atlas == 'AMBA':
        volume = np.array((528, 320, 456)) 
        posx, posy, posz = volume/2

    if atlas == 'WHS':
        volume = np.array((512, 1024, 512))
        posx, posy, posz = 256, 512, 256

    translated_volume = volume - translation_vector

    cor_linear_point = (
    (((translated_volume[0] / 2)) * cross[0]) + ((translated_volume[2] / 2) * cross[2])) + k
    cor_Y = -(cor_linear_point / cross[1])        
    #     cor_axis = ((translated_volume[0] / 2, depth, translated_volume[2] / 2))

    sag_linear_point = (
    ((translated_volume[1] / 2) * cross[1]) + ((translated_volume[2] / 2) * cross[2])) + k
    sag_X = -(sag_linear_point / cross[0])
    #     sag_axis = ((translated_volume[1] / 2, depth, translated_volume[2] / 2))

    horz_linear_point = (
        ((translated_volume[0] / 2) * cross[0]) + ((translated_volume[1] / 2) * cross[1])) + k
    horz_Z = -(horz_linear_point / cross[2])
    if plane_of_section is None:
        plane_of_section = np.argmin(np.abs((cor_Y - posy, sag_X - posx, horz_Z - posz)))
    choices = {'x':  sag_X, ' y': cor_Y,' z':  horz_Z }

    if plane_of_section == 0:
        axis = ((translated_volume[0] / 2, cor_Y, translated_volume[2] / 2))
        print(f'cor: {axis + translation_vector}')

        if direction == 'DV':
            linear_point = (
                ((translated_volume[0]) * cross[0]) + ((translated_volume[2] / 2) * cross[2])) + k
            Ypred = -(linear_point / cross[1])
            ##the way QNII rotates is but i prefer my way
        #             axis2 = ((translated_volume[0], cor_Y, translated_volume[2] / 2))
            axis2 = ((translated_volume[0], Ypred, translated_volume[2] / 2))

        ##this gives me the depth of a point directly beside the coronal center point
        if direction == 'ML':
            linear_point = (((translated_volume[0] / 2) * cross[0]) + ((translated_volume[2]) * cross[2])) + k 
            Ypred = -(linear_point / cross[1])
            axis2 = ((translated_volume[0] / 2, Ypred, translated_volume[2]))
        
        
        
        
    if plane_of_section==1:
        axis = ((sag_X, translated_volume[1] / 2, translated_volume[2] / 2))
        print(f'sag: {axis + translation_vector}')

        if direction == 'DV':
            linear_point = (
                ((translated_volume[1]) * cross[1]) + ((translated_volume[2] / 2) * cross[2])) + k
            Xpred = -(linear_point / cross[0])
            axis2 = ((Xpred, translated_volume[1], translated_volume[2] / 2))
        if direction == 'ML':
            linear_point = (
                ((translated_volume[1] / 2) * cross[1]) + ((translated_volume[2]) * cross[2])) + k
            Xpred = -(linear_point / cross[0])
            axis2 = ((Xpred, translated_volume[1] / 2, translated_volume[2]))
            
        
    if plane_of_section==2:
        axis = ((translated_volume[0] / 2,  translated_volume[1] / 2, horz_Z))
        print(f'horz: {axis + translation_vector}')
        if direction == 'DV':
            linear_point = (
                ((translated_volume[0]) * cross[0]) + ((translated_volume[1] / 2) * cross[1])) + k
            Zpred = -(linear_point / cross[2])
            axis2 = ((translated_volume[0], translated_volume[1] / 2, Zpred))


        if direction == 'ML':
            linear_point = (
                ((translated_volume[0] / 2) * cross[0]) + ((translated_volume[1]) * cross[1])) + k
            Zpred = -(linear_point / cross[2])
            axis2 = ((translated_volume[0] / 2, translated_volume[1], Zpred))
    axis_vector = np.array(axis) - np.array(axis2)
    return (axis_vector)



def rotate_section(section, degrees, direction, plane_of_section=None, atlas = 'AMBA'):
    '''
    Rotates a section
    :param section: the section to rotate
    :type section: :any:`numpy.ndarray`
    :param degrees: the degrees to rotate the section
    :type degrees: float
    :param direction: the direction of the rotation
    :type direction: string
    :param plane: the plane to rotate around
    :type plane: string
    :returns: the rotated section
    :rtype: :any:`numpy.ndarray`
    '''

    cross, k = find_plane_equation(section)

    # this looks redundant
    # if direction==ML:
    #   ML=get_angle(section.reshape(9),cross,k,direction=direction)
    section_points = section.copy()
    for i in range(3):
        section_points[i + 3] += section_points[i]
        section_points[i + 6] += section_points[i]

    points = section_points.reshape(3, 3)
    if atlas == 'WHS':
        translated_volume = np.array((489, 1024, 590))
        posx, posy, posz = 256, 512, 256
    if atlas == 'AMBA':
        translated_volume = np.array((528, 320, 456))
        posx, posy, posz = translated_volume/2

    cor_linear_point = (
        (((translated_volume[0] / 2)) * cross[0]) + ((translated_volume[2] / 2) * cross[2])) + k
    cor_Y = -(cor_linear_point / cross[1])        
#     cor_axis = ((translated_volume[0] / 2, depth, translated_volume[2] / 2))
    
    sag_linear_point = (
        ((translated_volume[1] / 2) * cross[1]) + ((translated_volume[2] / 2) * cross[2])) + k
    sag_X = -(sag_linear_point / cross[0])
#     sag_axis = ((translated_volume[1] / 2, depth, translated_volume[2] / 2))
    
    horz_linear_point = (
            ((translated_volume[0] / 2) * cross[0]) + ((translated_volume[1] / 2) * cross[1])) + k
    horz_Z = -(horz_linear_point / cross[2])
    if plane_of_section is None:
        plane_of_section = np.argmin(np.abs((cor_Y - posy, sag_X - posx, horz_Z - posz)))
#     midpoint = translated_volume/2
#     x = symbols('x')
#     expr = sum((x * cross + midpoint) * cross) - k
#     m = solve(expr)
#     translation_vector = np.array((m * cross + midpoint), dtype=np.float)
    if plane_of_section==0:
        
        translation_vector = (
            (translated_volume[0] / 2, cor_Y, translated_volume[2] / 2))
        
    if plane_of_section==1:
        translation_vector = (
            (sag_X, translated_volume[1] / 2, translated_volume[2] / 2))
        
    if plane_of_section==2:
        translation_vector = (
            (translated_volume[0] / 2, translated_volume[1] / 2, horz_Z))

    translated_points = points - translation_vector
    axis = get_axis(section, translation_vector, direction=direction, plane_of_section=plane_of_section)
    print(axis)
    rot_matrix = rotation_around_axis(axis, degrees)
    # perform rotation, centred on (0,0,0)
    rotated_translated_points = np.dot(translated_points, rot_matrix)
    # translate back to original geometric centre.
    rotated_points = rotated_translated_points + translation_vector
    rotated_points = rotated_points.reshape(9)
    for i in range(3):
        rotated_points[i + 3] -= rotated_points[i]
        rotated_points[i + 6] -= rotated_points[i]
    return (rotated_points)



def Section_adjust(section, direction, mean):
    cross, k = find_plane_equation(section)
    angle = get_angle(section, cross, k, direction)
    dif = angle-mean
    rot = (rotate_section(section, -dif, direction))
    return(rot)



import pandas as pd
import xml.etree.ElementTree as ET
from tqdm import tqdm
import re
import numpy as np


def indent(elem, level=0):
    i = "\n" + level * "  "
    if len(elem):
        if not elem.text or not elem.text.strip():
            elem.text = i + "  "
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
        for elem in elem:
            indent(elem, level + 1)
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i


# converts a pandas DataFrame to a quickNII compatible XML
def pd_to_quickNII(
    results,
    orientation="coronal",
    filename="Download",
    web=False,
    folder_name=None,
    aligner=None,
):
    # Replace the subfolder present in results if we are running for the web
    if web and folder_name:
        results["Filenames"] = (
            results["Filenames"]
            .str.replace(folder_name, "")
            .str.replace("\\", "")
            .str.replace("/", "")
        )
    # Get the total number of sections
    num_of_sections = results.shape[0]

    if not "section_ID" in results:
        section_numbers = np.arange((num_of_sections)) + 1
        if orientation == "coronal":
            results = results.sort_values("oy")
        if orientation == "sagittal":
            results = results.sort_values("ox")
        if orientation == "horizontal":
            results = results.sort_values("oz")
        results["section_ID"] = section_numbers
    else:
        results = results.sort_values("section_ID")
    results = results[
        [
            "Filenames",
            "ox",
            "oy",
            "oz",
            "ux",
            "uy",
            "uz",
            "vx",
            "vy",
            "vz",
            "section_ID",
            "width",
            "height",
        ]
    ]
    # Create the XML structure
    root = ET.Element("series")
    root.attrib["first"] = str(results["section_ID"].values)
    root.attrib["last"] = str(results["section_ID"].values)
    root.attrib["aligner"] = str(aligner)
    # Explicitly confirm all filenames are Strings
    results["Filenames"] = results["Filenames"].astype(str)
    # for each section append Oxyz, Uxyz and Vxyz parameters to the XML
    for i in range(num_of_sections):
        child = ET.SubElement(root, "slice")
        # this is the filename in our results file
        child.attrib["filename"] = results.iloc[i, 0]
        root.attrib["name"] = results.iloc[i, 0]  # so is this
        # Organise our coordinates
        ox, oy, oz, ux, uy, uz, vx, vy, vz = results.iloc[i, 1:10]
        # these next two values are the original file  size of the image, i need to fix this later.
        child.attrib["height"] = results.iloc[
            i,
        ]
        child.attrib["width"] = "-999"
        # Section number
        child.attrib["nr"] = str(results.iloc[i, 10])
        child.attrib["width"] = str(results.iloc[i, 11])
        child.attrib["height"] = str(results.iloc[i, 12])
        # writes Oxyz, Uxyz and Vxyz parameters to the XML in the correct format
        child.attrib["anchoring"] = (
            "ox="
            + str(ox)
            + "&oy="
            + str(oy)
            + "&oz="
            + str(oz)
            + "&ux="
            + str(ux)
            + "&uy="
            + str(uy)
            + "&uz="
            + str(uz)
            + "&vx="
            + str(vx)
            + "&vy="
            + str(vy)
            + "&vz="
            + str(vz)
        )
    indent(root)
    ET.ElementTree(root).write("{}.xml".format(filename))
    results.to_csv("{}.csv".format(filename))


# a useful script that converts a Quicknii XML to a csv file
# handy for training on human-aligned QuickNII files.


def XML_to_csv(xml):
    tree = ET.parse(str(xml))
    root = tree.getroot()
    count = 0
    df = pd.DataFrame()
    for i in root.findall("slice"):
        try:
            stringdata = str(i.attrib["anchoring"])
        except KeyError:
            continue
        df.loc[count, "Filenames"] = i.attrib["filename"]
        df.loc[count, "ox"] = re.search("ox=(.+?)&oy", stringdata).group(1)
        df.loc[count, "oy"] = re.search("oy=(.+?)&oz", stringdata).group(1)
        df.loc[count, "oz"] = re.search("oz=(.+?)&ux", stringdata).group(1)
        df.loc[count, "ux"] = re.search("ux=(.+?)&uy", stringdata).group(1)
        df.loc[count, "uy"] = re.search("uy=(.+?)&uz", stringdata).group(1)
        df.loc[count, "uz"] = re.search("uz=(.+?)&vx", stringdata).group(1)
        df.loc[count, "vx"] = re.search("vx=(.+?)&vy", stringdata).group(1)
        df.loc[count, "vy"] = re.search("vy=(.+?)&vz", stringdata).group(1)
        df.loc[count, "vz"] = re.search("vz=(.+?)$", stringdata).group(1)
        count += 1
    return df



from scyjava import jimport
from jpype import JImplements, JOverride
from jpype.types import JString, JDouble, JInt
import os

from src.abba_python.scijava_python_command import ScijavaCommand, ScijavaInput
import itk
import numpy as np
import tempfile

ExternalABBARegistrationPlugin = jimport('ch.epfl.biop.atlas.aligner.plugin.ExternalABBARegistrationPlugin')
AffineTransform3D = jimport('net.imglib2.realtransform.AffineTransform3D')
MultiSlicePositioner = jimport('ch.epfl.biop.atlas.aligner.MultiSlicePositioner')

Supplier = jimport('java.util.function.Supplier')

Command = jimport('org.scijava.command.Command')
JPlugin = jimport('org.scijava.plugin.Plugin')
Parameter = jimport('org.scijava.plugin.Parameter')
ImagePlus = jimport('ij.ImagePlus')

SimpleABBARegistrationPlugin = jimport('ch.epfl.biop.atlas.aligner.plugin.SimpleABBARegistrationPlugin')

ElastixTransform = jimport('itc.transforms.elastix.ElastixTransform')
ElastixEuler2DToAffineTransform3D = jimport('itc.converters.ElastixEuler2DToAffineTransform3D')
File = jimport('java.io.File')

SourcesChannelsSelect = jimport('ch.epfl.biop.sourceandconverter.processor.SourcesChannelsSelect')
HashMap = jimport('java.util.HashMap')
SimpleRegistrationWrapper = jimport('ch.epfl.biop.atlas.aligner.plugin.SimpleRegistrationWrapper')


@JImplements(SimpleABBARegistrationPlugin)
class ITKRigidRegistration(object):

    def __init__(self, ij):
        print('init SimpleRotateAffineRegistration')
        self.ij = ij
        pass

    # @return Sampling required for the registration, in micrometer (double)
    #
    @JOverride
    def getVoxelSizeInMicron(self):
        return JDouble(40)

    # Is called before registration to pass any extra registration parameter
    # argument. Passed as a dictionary of String to preserve serialization
    # capability.
    # param parameters dictionary of parameters (Map<String, String>)
    @JOverride
    def setRegistrationParameters(self, parameters):
        self.parameters = parameters

    # param fixed image (ImagePlus)
    # param moving image (ImagePlus)
    # param fixed mask (ImagePlus)
    # param moving mask (ImagePlus)
    # return the transform, result of the registration, (InvertibleRealTransform)
    # going from fixed to moving coordinates, in pixels
    @JOverride
    def register(self, fixed, moving, fixedMask, movingMask):
        # fixed.show()
        # moving.show()
        # ij.py.show(fixed)
        # ij.py.show(moving)

        # global parameter_object
        parameter_object = itk.ParameterObject.New()
        default_rigid_parameter_map = parameter_object.GetDefaultParameterMap('rigid')
        parameter_object.AddParameterMap(default_rigid_parameter_map)

        # global fixed_py
        # global moving_py
        fixed_py = self.ij.py.from_java(fixed)
        moving_py = self.ij.py.from_java(moving)
        fixed_py = fixed_py.to_numpy().astype(np.float32)
        moving_py = moving_py.to_numpy().astype(np.float32)

        # Call registration function
        with tempfile.TemporaryDirectory() as temp_dir:
            # ... do something with temp_dir
            itk.elastix_registration_method(
                fixed_py, moving_py,
                output_directory=temp_dir,
                parameter_object=parameter_object,
                log_to_console=False)
            output_path = os.path.join(temp_dir, 'TransformParameters.0.txt')
            et = ElastixTransform.load(File(output_path))
            transform = ElastixEuler2DToAffineTransform3D.convert(et)

        # temp dir automatically cleaned up when context exited
        return transform


@JImplements(Supplier)
class ITKRigidRegistrationSupplier(object):

    def __init__(self, ij):
        self.ij = ij
        pass

    @JOverride
    def get(self):
        return SimpleRegistrationWrapper(JString('ITKRigidRegistration'),
                                         ITKRigidRegistration(self.ij))


def add_abba_itk_registrations(ij):
    MultiSlicePositioner.registerRegistrationPlugin('ITKRigidRegistration',
                                                    ITKRigidRegistrationSupplier(ij))

    @ScijavaCommand(context=ij.context(),  # ij context needed
                    name='ITK - Rigid Registration')
    @ScijavaInput('fixed_channel', JInt,
                  label='Atlas channel index:', description='')
    @ScijavaInput('moving_channel', JInt,
                  label='Section channel index', description='')
    @ScijavaInput('mp', MultiSlicePositioner,
                  label='Section channel index', description='')
    class SimpleRotateAffineRegistrationCommand:
        def run(self):
            params = HashMap()
            self.mp.registerSelectedSlices('ITKRigidRegistration',
                                           SourcesChannelsSelect(self.fixed_channel),
                                           SourcesChannelsSelect(self.moving_channel),
                                           params)

    MultiSlicePositioner.registerRegistrationPluginUI('ITKRigidRegistration',
                                                      'ITK - Rigid Registration');



##very janky way to solve relative import problem
import os
from pathlib import Path
# store ini paths
ini_path = os.getcwd()
##set path to be the DeepSlice directory
path = str(Path(__file__).parent)
##testing automated tests
os.chdir(path)
print(path)
import warnings
from tensorflow.keras.applications.xception import Xception
from tensorflow.keras import Sequential
from tensorflow.keras.layers import Dense
from skimage import color
from tensorflow.keras.preprocessing.image import ImageDataGenerator
from src.abba_python.deepslice.utilities.QuickNII_functions import pd_to_quickNII, XML_to_csv
from src.abba_python.deepslice.utilities import plane_alignment
import pandas as pd
import numpy as np
from sklearn.linear_model import HuberRegressor
import re
import imghdr
import struct

# restore path
os.chdir(ini_path)

def get_image_size(fname):
    # https://stackoverflow.com/questions/8032642/how-to-obtain-image-size-using-standard-python-class-without-using-external-lib
    """Determine the image type of fhandle and return its size.
    from draco"""
    with open(fname, "rb") as fhandle:
        head = fhandle.read(24)
        if len(head) != 24:
            raise Exception("Invalid header")

        ext = imghdr.what(fname)
        if imghdr.what(fname) == "png":
            check = struct.unpack(">i", head[4:8])[0]
            if check != 0x0D0A1A0A:
                raise Exception("png checksum failed")
            width, height = struct.unpack(">ii", head[16:24])
        elif imghdr.what(fname) == "gif":
            width, height = struct.unpack("<HH", head[6:10])
        elif imghdr.what(fname) == "jpeg":
            fhandle.seek(0)  # Read 0xff next
            size = 2
            ftype = 0
            while not 0xC0 <= ftype <= 0xCF:
                fhandle.seek(size, 1)
                byte = fhandle.read(1)
                while ord(byte) == 0xFF:
                    byte = fhandle.read(1)
                ftype = ord(byte)
                size = struct.unpack(">H", fhandle.read(2))[0] - 2
            # We are at a SOFn block
            fhandle.seek(1, 1)  # Skip `precision' byte.
            height, width = struct.unpack(">HH", fhandle.read(4))
        else:
            raise Exception(f"Invalid filetype: {head}")
        return width, height


def ideal_thickness(results, depth, detect_bad_sections=False):
    number_spacing = np.float64(results["section_ID"].values[1:]) - np.float64(
        results["section_ID"].values[:-1]
    )
    depth_spacing = np.array(depth)[:-1] - np.array(depth)[1:]
    norm_depth = depth_spacing / number_spacing

    if detect_bad_sections == True:
        predicted_thickness = np.mean(norm_depth)
        thickness_variability = np.std(norm_depth)
        good_indexes = (
            norm_depth < (predicted_thickness + (thickness_variability * 3))
        ) & (norm_depth > (predicted_thickness - (thickness_variability * 3)))
        print(
            f"sections {results.Filenames[1:][~good_indexes]} appeared to be outliers and were not incorporated into the spacing analysis"
        )
        norm_depth = norm_depth[good_indexes]
    predicted_thickness = np.mean(norm_depth) * 25
    thickness_variability = np.std(norm_depth) * 25

    print(
        "Your sections appear to be sectioned at {0:.4f} micron thickness".format(
            np.abs(predicted_thickness)
        )
    )
    print(
        "the variability of thickness is {0:.4f} microns".format(thickness_variability)
    )
    return predicted_thickness


def ideal_spacing(
    pred_oy, section_numbers, section_thickness_um, bad_section_indexes=None
):
    if bad_section_indexes is None:
        bad_section_indexes = np.array([False] * len(pred_oy))

    pred_oy = np.float64(pred_oy[~bad_section_indexes])
    section_numbers = np.float64(section_numbers.values)
    section_thickness_um = np.float64(section_thickness_um)
    pred_um = pred_oy * 25
    section_um = section_numbers * section_thickness_um
    print(f"section_um: {pred_um/25}")
    avg_dist = np.mean(pred_um - section_um[~bad_section_indexes])
    print("ideal: ", (section_um + avg_dist) / 25)
    print("pred_UM: ")
    print(f"average distance: {avg_dist}")
    # print((pred_um-section_um) - avg_dist)
    return ((section_um + avg_dist)) / 25


def calculate_brain_center_depth(section):
    cross, k = plane_alignment.find_plane_equation(section)
    translated_volume = np.array((456, 0, 320))
    linear_point = (
        ((translated_volume[0] / 2) * cross[0])
        + ((translated_volume[2] / 2) * cross[2])
    ) + k
    depth = -(linear_point / cross[1])
    return depth

def get_deepslice_path():
    return path

class DeepSlice:
    def __init__(self, web=False, folder_name=None):
        self.web = web
        self.folder_name = folder_name

        self.image_dir = None
        self.columns = ["ox", "oy", "oz", "ux", "uy", "uz", "vx", "vy", "vz"]

    def init_model(self, DS_weights, xception_weights):
        # Download Xception architecture with weights pretrained on imagenet
        DenseModel = Xception(include_top=True, weights=xception_weights)
        # remove the Dense Softmax layer and average pooling layer from the pretrained model
        DenseModel._layers.pop()
        DenseModel._layers.pop()
        # Build Deepslice
        model = Sequential()
        model.add(DenseModel)
        # we tested various sizes for these last two layers but consistently found that 256 performed best for some unknown reason.
        # theoretically larger layers should be better able to fit the training set but this is not what we saw.
        model.add(Dense(256, activation="relu"))
        model.add(Dense(256, activation="relu"))
        # as we are predicting continuous values, here we define 9 output neurons with linear activation functions,
        # each corresponding to one of the QuickNII alignment variables Oxyz, Uxyz, Vxyz.
        model.add(Dense(9, activation="linear"))
        if DS_weights != None:
            # load weights
            model.load_weights(DS_weights)
        return model

    def Build(
        self,
        DS_weights=path + "/NN_weights/Allen_Mixed_Best.h5",
        xception_weights=path
        + "/NN_weights/xception_weights_tf_dim_ordering_tf_kernels.h5",
        wise_weights=path + "/NN_weights/Synthetic_data_final.hdf5",
        species="mouse",
    ):
        self.wise_weights = wise_weights
        self.DS_weights = DS_weights
        self.rat_weights = path + "/NN_weights/rat_mixed_4056.h5"
        if species.lower() == "rat":
            self.model = self.init_model(
                DS_weights=self.rat_weights, xception_weights=xception_weights
            )
        if species.lower() == "mouse":
            self.model = self.init_model(
                DS_weights=self.DS_weights, xception_weights=xception_weights
            )
        self.species = species

    def gray_scale(self, img):
        # Downsamples images too 299 x 299
        # converts images to grayscale
        img = color.rgb2gray(img).reshape(299, 299, 1)
        return img

    def predict(self, image_dir, prop_angles=True, huber=False, wise=False):  # input
        # define_image_generator
        self.Image_generator = ImageDataGenerator(
            preprocessing_function=self.gray_scale, samplewise_std_normalization=True
        ).flow_from_directory(
            image_dir,
            target_size=(299, 299),
            batch_size=1,
            color_mode="rgb",
            shuffle=False,
        )
        self.image_dir = image_dir
        # reset the image generator to ensure it starts from the first image
        self.Image_generator.reset()
        # feed images to the model and store the predicted parameters
        preds = self.model.predict(
            self.Image_generator,
            steps=self.Image_generator.n // self.Image_generator.batch_size,
            verbose=1,
        )
        # convert the parameter values to floating point digits
        preds = preds.astype(float)
        if wise:
            self.Image_generator.reset()
            self.model.load_weights(self.wise_weights)
            wise_preds = self.model.predict(
                self.Image_generator,
                steps=self.Image_generator.n // self.Image_generator.batch_size,
                verbose=1,
            )

            preds = np.mean((preds, wise_preds), axis=0)
            self.model.load_weights(self.DS_weights)

        # define the column names
        # create a pandas DataFrame of the parameter values
        results = pd.DataFrame(preds, columns=self.columns)
        # insert the section filenames into the pandas DataFrame
        results["Filenames"] = self.Image_generator.filenames[: results.shape[0]]
        ordered_cols = ["Filenames"] + self.columns
        self.results = results[ordered_cols]  # To get the same column order
        if prop_angles:
            self.propagate_angles(huber)

    def reorder_indexes(self, ascending):
        # reorders values so they are in the same order as the indexes with minimal swaps
        self.results.oy = self.results.oy.sort_values(ascending=ascending).values

    def even_spacing(
        self,
        section_thickness_um=None,
        no_correction=False,
        order_only=False,
        bad_sections=[],
        detect_bad_sections=False,
        ignore_final_chars=0,
    ):
        print(
            "Section Numbers must have been included as the last three digits of the Filename"
        )
        ###This function takes a dataset with section numbers and spaces those sections based on their numbers
        section_numbers = []
        depth = []
        count = 1
        for Filename in (
            self.results.Filenames.str.split("\\", expand=True).iloc[:, -1].values
        ):
            if ignore_final_chars > 0:
                temp_filename = Filename[:-ignore_final_chars]
            else:
                temp_filename = Filename
            ##this removes all non-numeric characters
            section_number = re.sub("[^0-9]", "", temp_filename)
            ###this gets the three numbers closest to the end
            section_number = section_number[-3:]
            ind = [Filename in i for i in self.results.Filenames.values]
            d = self.results[ind]
            d = calculate_brain_center_depth(d[self.columns].values[0])
            # print(f"Filename: {Filename} section_number: {section_number} depth: {d}")

            ###find the first appearancex of the specified pattern
            ###remove non-numeric characters
            if len(section_number) < 3:
                warnings.warn(
                    f'could not find three digit section number for file "{Filename}", it should be the last three digits of the filenames.'
                )
            if len(section_number) == 0:
                warnings.warn(
                    f'could not find any section number for file "{Filename}\, using {count} instead'
                )
                section_numbers.append(count)
            count += 1
            section_numbers.append(section_number)

        self.results["section_ID"] = section_numbers

        self.results.section_ID = self.results.section_ID.astype(np.float64)
        self.results = self.results.sort_values(
            "section_ID", ascending=True
        ).reset_index(drop=True)

        print(self.results)

        depth = []
        for section in self.results[self.columns].values:
            depth.append((calculate_brain_center_depth(section)))
        depth = np.array(depth)

        self.results["depth"] = pd.Series(depth)
        if len(bad_sections) > 0:
            bad_section_indexes = np.sum(
                [self.results.Filenames.str.contains(bs) for bs in bad_sections],
                axis=0,
                dtype=bool,
            )
            print(
                f" we found {np.sum(bad_section_indexes)} out of {len(bad_sections)} bad sections"
            )
            estimate_thickness = ideal_thickness(
                self.results[~bad_section_indexes],
                depth[~bad_section_indexes],
                detect_bad_sections=detect_bad_sections,
            )
        else:
            estimate_thickness = ideal_thickness(
                self.results, depth, detect_bad_sections=detect_bad_sections
            )

        print("\n", estimate_thickness, "\n")
        print("new version")

        if estimate_thickness > 0:
            print("the sections are numbered rostrocaudaly")
            self.reorder_indexes(ascending=False)

        else:
            print("the sections are not numbered rostrocaudaly")
            self.reorder_indexes(ascending=True)

            if section_thickness_um is not None:
                section_thickness_um *= -1

        if no_correction or order_only:
            return

        depth = []
        for section in self.results[self.columns].values:
            depth.append((calculate_brain_center_depth(section)))
        depth = np.array(depth)

        if len(bad_sections) > 0:
            bad_section_indexes = np.sum(
                [self.results.Filenames.str.contains(bs) for bs in bad_sections],
                axis=0,
                dtype=bool,
            )
            print(
                f" we found {np.sum(bad_section_indexes)} out of {len(bad_sections)} bad sections"
            )
            estimate_thickness = ideal_thickness(
                self.results[~bad_section_indexes],
                depth[~bad_section_indexes],
                detect_bad_sections=detect_bad_sections,
            )
        else:
            estimate_thickness = ideal_thickness(
                self.results, depth, detect_bad_sections=detect_bad_sections
            )

        if section_thickness_um is None:
            section_thickness_um = -estimate_thickness
        print(len(depth))
        if len(bad_sections) > 0:
            ideal = ideal_spacing(
                depth,
                self.results["section_ID"],
                section_thickness_um,
                bad_section_indexes=bad_section_indexes,
            )
        else:
            ideal = ideal_spacing(
                depth, self.results["section_ID"], section_thickness_um
            )

        self.results.oy -= depth - ideal
        depth = []
        for section in self.results[self.columns].values:
            depth.append((calculate_brain_center_depth(section)))
        self.results.section_ID = np.abs(self.results.section_ID)

    def load_QuickNII(self, filename):
        self.results = XML_to_csv(filename)
        self.results[self.columns] = self.results[self.columns].astype(np.float64)

    def set_angles(self, DV=None, ML=None):
        rotated_sections = []
        count = 0
        for section in self.results.iterrows():
            m = section[1][
                ["ox", "oy", "oz", "ux", "uy", "uz", "vx", "vy", "vz"]
            ].values.astype(np.float64)
            cross, k = plane_alignment.find_plane_equation(m)
            old_DV = plane_alignment.get_angle(m, cross, k, "DV")
            old_ML = plane_alignment.get_angle(m, cross, k, "ML")
            section = section[1][self.columns].values
            original_depth = calculate_brain_center_depth(section)
            for i in range(10):
                cross, k = plane_alignment.find_plane_equation(section)
                if DV is not None:
                    section = plane_alignment.Section_adjust(
                        section, mean=DV, direction="DV"
                    )
                else:
                    section = plane_alignment.Section_adjust(
                        section, mean=old_DV, direction="DV"
                    )
                if ML is not None:
                    section = plane_alignment.Section_adjust(
                        section, mean=ML, direction="ML"
                    )
                else:
                    section = plane_alignment.Section_adjust(
                        section, mean=old_ML, direction="DV"
                    )
            rotated_depth = calculate_brain_center_depth(section)
            movement = rotated_depth - original_depth
            # section[1] -= movement
            rotated_sections.append(section)
            cross, k = plane_alignment.find_plane_equation(section)
            final_depth = calculate_brain_center_depth(section)
            # print(" original: {} \n rotated {} \n corrected {} \n".format(original_depth, rotated_depth, final_depth))

            count += 1
        self.results[self.columns] = rotated_sections
        # insert the section filenames into the pandas DataFrame

    def propagate_angles(self, huber=True):
        DV = []
        ML = []
        oy = []
        for prediction in self.results.iterrows():
            m = prediction[1][
                ["ox", "oy", "oz", "ux", "uy", "uz", "vx", "vy", "vz"]
            ].values.astype(np.float64)
            oy.append(m[1])
            cross, k = plane_alignment.find_plane_equation(m)
            DV.append(plane_alignment.get_angle(m, cross, k, "DV"))
            ML.append(plane_alignment.get_angle(m, cross, k, "ML"))
        if huber == True:
            oy = np.array(oy).reshape(-1, 1)
            # we use a huberised linear regressor as it is more robust to outliers
            huber_regressor = HuberRegressor().fit(oy, DV)
            # for our predictions we multiple the depth by the coefficient and add the y intercept
            prop_DV = (huber_regressor.coef_ * oy) + huber_regressor.intercept_
            huber_regressor = HuberRegressor().fit(oy, ML)
            prop_ML = (huber_regressor.coef_ * oy) + huber_regressor.intercept_
        else:
            length = len(DV)
            weights = plane_alignment.make_gaussian_weights(0, 528)
            oy = np.array(oy)
            oy[oy < 0] = 0
            oy[oy > 527] = 527
            weights = [weights[int(y)] for y in oy]
            # DV = sorted(DV, key=abs)[int(length*0.75):]
            # ML = sorted(ML, key=abs)[int(length*0.75):]
            print(weights)
            prop_DV = [np.average(DV, weights=np.array(weights))] * length
            prop_ML = [np.average(ML, weights=np.array(weights))] * length

        rotated_sections = []
        count = 0
        for section in self.results.iterrows():
            section = section[1][self.columns].values
            original_depth = calculate_brain_center_depth(section)
            for i in range(10):
                cross, k = plane_alignment.find_plane_equation(section)

                section = plane_alignment.Section_adjust(
                    section, mean=prop_DV[count], direction="DV"
                )

                section = plane_alignment.Section_adjust(
                    section, mean=prop_ML[count], direction="ML"
                )
            rotated_depth = calculate_brain_center_depth(section)
            movement = rotated_depth - original_depth
            # section[1] -= movement
            rotated_sections.append(section)
            cross, k = plane_alignment.find_plane_equation(section)
            final_depth = calculate_brain_center_depth(section)
            # print(" original: {} \n rotated {} \n corrected {} \n".format(original_depth, rotated_depth, final_depth))

            count += 1
        results = pd.DataFrame(rotated_sections, columns=self.columns)
        # insert the section filenames into the pandas DataFrame
        results["Filenames"] = self.Image_generator.filenames[: results.shape[0]]
        ordered_cols = ["Filenames"] + self.columns
        self.results = results[ordered_cols]  # To get the same column order

    def Save_Results(self, filename):
        if "section_ID" in self.results:
            section_numbers = np.abs(self.results["section_ID"])
        else:
            section_numbers = None
        widths, heights = [], []
        if self.image_dir is not None:
            for file in self.results.Filenames:
                width, height = get_image_size(self.image_dir + os.path.sep + file)
                widths.append(width)
                heights.append(height)
        else:
            for file in self.results.Filenames:
                width, height = get_image_size(file)
                widths.append(width)
                heights.append(height)
        self.results["width"] = widths
        self.results["height"] = heights
        pd_to_quickNII(
            results=self.results,
            orientation="coronal",
            filename=str(filename),
            web=self.web,
            folder_name=self.folder_name,
            aligner="DeepSlice_ver_3.0_python",
        )

